-module(lazy).
-compile(export_all).
-compile({no_auto_import,[tl/1,hd/1]}).
-author('hanak@emt.bme.hu, kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2020-11-17 07:14:09 +0100 (k, 17 nov 2020) $$').

-type lazy_list() :: nil | {any(), fun(() -> lazy_list())}.
% Ez a lista félig lusta: a fej mindig kiértékelődik, a farok lusta.
% A teljesen lusta lista kicsit komplikáltabb, mert nem tudni, hogy
% üres vagy sem.

-spec hd(Ys::lazy_list()) -> H::any().
% Az Ys lusta farkú lista feje H.
hd({H,_Ts}) ->
    H.

-spec tl(Ys::lazy_list()) -> Ts::lazy_list().
% Az Ys lusta farkú lista farka Ts.
tl({_H,Ts}) ->
    Ts().

-spec infseq(N::integer()) -> Ys::lazy_list().
% Egész számok N-nel kezdődő, egyesével növekedő, nem korlátos
% sorozata az Ys lusta farkú lista.
infseq(N) ->
    {N, fun() -> infseq(N+1) end}.

-spec seq(M::integer(), N::integer()) -> Ys::lazy_list().
% Egész számok M-től N-ig tartó, egyesével növekedő sorozata
% az Ys lusta farkú lista.
seq(M, N) when M =< N ->
    {M, fun() -> seq(M+1, N) end};
seq(_, _) ->
    nil.

-spec cons(Ls::list()) -> Ys::lazy_list().
% Az Ls lista elemeiből álló lusta farkú lista Ys.
cons([]) ->
    nil;
cons([H|Ts]) ->
    {H, fun() -> cons(Ts) end}.

-spec take(Ys::lazy_list(), N::integer()) -> Ls::list().
% Az Ys lusta farkú lista első N eleméből álló lista Ls.
take(_,0) ->
    [];
take(nil,_) ->
    [];
take({H,_}, 1) ->  % optimalizáció: hogy ne kelljen a farkát kiértékelni T()
    [H];
take({H,Ts}, N) ->
    [H|take(Ts(), N-1)].

% Szószátyár take nyomkövetéshez.
-spec take2(Ys::lazy_list(), N::integer()) -> Ls::list().
% Az Ys lusta farkú lista első N eleméből álló lista Ls.
take2(L,0) ->
    io:format("  Hívás (1. klóz): take(~w, 0)~n", [L]),
    [];
take2(nil,N) ->
    io:format("  Hívás (2. klóz): take(~w, ~w)~n", [nil, N]),
    [];
take2({H,T},1) ->
    io:format("  Hívás (3. klóz): take(~w, ~w)~n", [{H,T}, 1]),
    [H];
take2({H,T},N) ->
    io:format("  Hívás (4. klóz): take(~w, ~w)~n", [{H,T}, N]),
    [H|take2(T(), N-1)].

% sum/1 és sum/2 csak véges lusta farkú listákra!
% Az Ys véges lusta farkú számlista elemeinek összege N
-spec sum(Ys::lazy_list()) -> N::number().
sum(L) ->
    sum(L, 0).

% Az Ys véges lusta farkú számlista elemeinek és az A számnak
% az összege N.
-spec sum(Ys::lazy_list(), A::number()) -> N::number().
sum(nil, A) ->
    A;
sum({H,T}, A) ->
    sum(T(), H+A).

-spec map(F::fun((any()) -> any()), Ys::lazy_list()) -> Zs::lazy_list().
% Az Ys lusta farkú lista elemeiből az F alkalmazásával előálló lusta
% farkú lista Zs.
map(_, nil) ->
    nil;
map(F, {H,Ts}) ->
    {F(H), fun() -> map(F, Ts()) end}.

-spec filter(P::fun((any()) -> boolean()), Ys::lazy_list()) -> Zs::lazy_list().
% Az Ys lusta farkú lista elemei közül a P predikátummal kiválogatott elemekből
% álló lusta farkú lista Zs.
% Kicsit mohó, az eredménylista fejéig kiértékeli a listát.
filter(_, nil) ->
    nil;
filter(P, {H,Ts}) ->
    case P(H) of
        true ->
            {H, fun() -> filter(P, Ts()) end};
        false ->
            filter(P, Ts()) % Megkeressük az eredmény fejét.
    end.

-spec append(Xs::lazy_list(), Ys::lazy_list()) -> Zs::lazy_list().
% Az Xs lusta farkú lista Ys (ugyancsak lusta farkú lista) elé fűzésével előálló
% lusta farkú lista Zs. Ha Xs nem véges, az összefűzés felesleges, az Ys elemei
% elérhetetlenek lesznek.
append(nil, Ys) ->
    Ys;
append({H,Xs}, Ys) ->
    {H, fun() -> append(Xs(), Ys) end}.

% Szószátyár append nyomkövetéshez.
-spec append2(Xs::lazy_list(), Ys::lazy_list()) -> Zs::lazy_list().
% Az Xs lusta farkú lista Ys (ugyancsak lusta farkú lista) elé fűzésével előálló
% lusta farkú lista Zs. Ha Xs nem véges, az összefűzés felesleges, az Ys elemei
% elérhetetlenek lesznek.
append2(nil, Ys) ->
    io:format("  Hívás (1. klóz): append([], ~w)~n", [(Ys)]),
    Ys;
append2({H,Ts}, Ys) ->
    io:format("  Hívás (2. klóz): append(~w, ~w)~n", [([H|Ts]), (Ys)]),
    {H, fun() -> append2(Ts(), Ys) end}.

-spec fibs(Curr::integer(), Next::integer()) -> Ys::lazy_list().
% Ys egy olyan általánosított Fibonacci sorozat (lusta farkú lista),
% amelynek első két tagja Curr és Next.

fibs(Curr, Next) ->
    {Curr, fun() -> fibs(Next, Curr+Next) end}.

-spec fib(N::integer()) -> F::integer().
% Az N-edik Fibonacci-szám F.
fib(N) ->
    nth(fibs(0,1), N+1).

-spec sift(Prime::integer(), Ys::lazy_list()) -> Zs::lazy_list().
% Az Ys lusta farkú lista azon elemeinek listája a Zs lusta farkú
% lista, melyek nem oszthatók Prime-mal.
sift(Prime, Ys) ->
    filter(fun(N) -> N rem Prime =/= 0 end, Ys).

-spec sieve(Ys::lazy_list()) -> Zs::lazy_list().
% Az Zs lusta farkú lista az Ys nem véges lusta farkú lista
% "szitáltja" (üres listára hibát jelez).
sieve({H,Ts}) ->
    {H, fun() -> sieve(sift(H, Ts())) end}.

-spec qsort(Ys::lazy_list()) -> Zs::lazy_list().
% Az Ys lusta farkú lista rendezett változata a Zs lusta
% farkú lista.
% Az append alkalmazása: ha csak a lista elejére kíváncsi
% a felhasználó, csak a lista elejét rendezi.
qsort(nil) ->
    nil;
qsort({Pivot, Ys}) ->
    append(qsort(filter(fun(Y) -> Y < Pivot end, Ys())),
           {Pivot, fun() -> qsort(filter(fun(Y) -> Y >= Pivot end, Ys())) end}).

% Szószátyár qsort nyomkövetéshez.
-spec qsort2(Ys::lazy_list()) -> Zs::lazy_list().
% Az Ys lusta farkú lista rendezett változata a Zs lusta
% farkú lista.
qsort2(nil) ->
    nil;
qsort2({Pivot,Ys}) ->
    io:format(  "Hívás: qsort2(~w)~n", [take({Pivot,Ys}, 100)]),
    Low = fun(Y) -> Y < Pivot end,
    High = fun(Y) -> Y >= Pivot end,
    append2(qsort2(filter(Low, Ys())),
           {Pivot, fun() -> qsort2(filter(High, Ys())) end}).

-spec nth(Ys::lazy_list(), N::integer()) -> Y::any().
% Az Ys lusta lista N-edik eleme Y (számozás 1-től).
nth({H,_Ts}, 1) ->
    H;
nth({_H,Ts}, N) ->
    nth(Ts(), N-1).

-spec nth2(Ys::lazy_list(), N::integer()) -> {ok, Y::any()} | error.
% Az Ys lusta lista N-edik eleme Y (számozás 1-től). A visszatérési
% érték az 'error' atom, ha az Ys lista üres, vagy ha N < 1.
nth2({H,_Ts}, 1) ->
    {ok, H};
nth2({_H,Ts}, N) when N > 1 ->
    nth2(Ts(), N-1);
nth2(_, _) ->
    error.

test(1) ->
    Ys = {1, fun() -> {4, fun() -> {9,fun() -> nil end} end} end},
    [
     sum(seq(1,100000)) =:= 5000050000,
     sum(cons(lists:seq(1,100000))) =:= 5000050000,
     take(Ys, 2) =:= [1,4],
     take(Ys, 3) =:= [1,4,9],
     take(cons([1,4,9]), 4) =:= [1,4,9],
     take(seq(1,3),10) =:= [1,2,3],
     take(infseq(1),10) =:= [1,2,3,4,5,6,7,8,9,10],
     hd(tl(tl(infseq(1)))) =:= 3,
     take(tl(tl(infseq(1))),1) =:= [3],
     take(fibs(0,1),10+1) =:= [0,1,1,2,3,5,8,13,21,34,55],
     nth(fibs(0,1), 100+1) =:= 354224848179261915075,
     fib(100) =:= 354224848179261915075,
     nth2(fibs(0,1), 100+1) =:= {ok,354224848179261915075},
     nth2(nil, 5) =:= error,
     nth2(fibs(0,1), 0) =:= error,
     take(map(fun(X) -> X rem 2 =:= 0 end, infseq(3)), 3) =:= [false,true,false],
     take(filter(fun(X) -> X rem 2 =:= 0 end, infseq(3)), 3) =:= [4,6,8],
     take(filter(fun(X) -> X rem 2 =:= 1 end, seq(3,6)), 3) =:= [3,5],
     [take(append(cons("ab"),cons("12")), I) || I <- lists:seq(0,4)] =:=
	 [[],"a","ab","ab1","ab12"],
     take(sieve(infseq(2)), 10) =:= [2,3,5,7,11,13,17,19,23,29]
    ];
test(2) ->
    [
     take(qsort(cons([5,3,6,8,1,7])), 2) =:= [1,3],
     take(qsort(cons([5,3,6,8,1,7])), 5) =:= [1,3,5,6,7],
     io:format("~nTeszthívás: ~s~n", ["take(qsort2(cons([5,3,6,8,1,7])), 2)"]),
     take(qsort2(cons([5,3,6,8,1,7])), 2) =:= [1,3],
     io:format("~nTeszthívás: ~s~n", ["take(qsort2(cons([5,3,6,8,1,7])), 5)"]),
     take(qsort2(cons([5,3,6,8,1,7])), 5) =:= [1,3,5,6,7]
    ].

-spec print_mem_usage(string()) -> {integer(), non_neg_integer()}.
% Durva becslés: kiírja a függvényhívás előtt és után használt memóriát.
print_mem_usage(Msg) ->
    {memory,M1} = erlang:process_info(self(), memory),
    {_,Time} = statistics(runtime),
    io:format("~16s: ~3w M  |  ~.3f s~n", [Msg, M = M1 div 1024 div 1024, Time / 1000]),
    {M,Time}.

memtest(1) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    sum(seq(1,10000000)),
    print_mem_usage("seq o sum");
memtest(2) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    L = lists:seq(1,10000000),
    print_mem_usage("lists:seq"),
    lists:sum(L),
    print_mem_usage("lists:sum").
